
from enum import Enum

class ClientType(str, Enum):
    EXTERNAL = "EXTERNAL"
    USER = "USER"


class DataPointType(str, Enum):
    GRAPHQL = "GRAPHQL"
    REST = "REST"



class TYPENAMES(str, Enum):
    MODELPORTTYPE = "ModelPortType"


from bergen.types.node.inputs import Inputs, Outputs
from bergen.types.manager import ModelManager
from bergen.extenders.node import NodeExtender
from bergen.extenders.user import UserPrettifier
from bergen.extenders.port import PortExtender
from bergen.types.object import ArnheimObject
from bergen.types.model import ArnheimModel
from bergen.delayed import CREATE_NODE_MUTATION, NODE_FILTER_QUERY, NODE_QUERY
from enum import Enum
from typing import  Any, ForwardRef, Generic, List, Optional, Type, TypeVar


Node = ForwardRef("Node")
User = ForwardRef("User")
DataModel = ForwardRef("DataModel")

class Avatar(ArnheimObject):
    user: Optional['User']
    avatar: Optional[str]



class User(UserPrettifier, ArnheimModel):
    id: Optional[int]
    username: Optional[str]
    firstName: Optional[str]
    lastName: Optional[str]
    avatar: Optional[Avatar]

    class Meta:
        identifier = "user"
        get = NODE_QUERY



class DataPoint(ArnheimModel):
    type: Optional[str]
    name: Optional[str]
    host: Optional[str]
    port: Optional[int]
    url: Optional[str]
    models: Optional[List[DataModel]]

    class Meta:
        identifier = "datapoint"


class DataModel(ArnheimModel):
    identifier: Optional[str]
    extenders: Optional[List[str]]
    point: Optional[DataPoint]

    class Meta:
        identifier = "datamodel"


class Transcript(ArnheimObject):
    array: Optional[Any]
    extensions: Optional[Any]
    models: Optional[List[DataModel]]
    points: Optional[List[DataPoint]]
    user: Optional[User]


class PortType(PortExtender, ArnheimObject):
    required: Optional[bool]
    key: Optional[str]
    identifier: Optional[str] # Only for our friends the Models


class NodeManager(ModelManager['Node']):


    def get(self, ward=None, **kwargs) -> 'Node':
        return NODE_QUERY(self.model).run(ward=ward, variables=kwargs)


    def get_or_create(self, inputs: Type[Inputs] = None, outputs: Type[Outputs] = None , **kwargs) -> 'Node':
        
        parsed_inputs = inputs.serialized
        parsed_outputs = outputs.serialized
        
        node = CREATE_NODE_MUTATION(self.model).run(variables={
            "inputs" : parsed_inputs,
            "outputs": parsed_outputs,
            **kwargs

        })
        return node



class Node(NodeExtender, ArnheimModel):
    __slots__ = ("_pod","_provisionhandler")


    id: Optional[int]
    name: Optional[str]
    package: Optional[str]
    inputs: Optional[List[PortType]]
    outputs: Optional[List[PortType]]

    objects = NodeManager()

    class Meta:
        identifier = "node"
        filter = NODE_FILTER_QUERY


class Volunteer(ArnheimModel):
    id: int
    identifier: str
    node: Node

    class Meta:
        identifier = "volunteer"

class Template(ArnheimModel):
    node: Optional[Node]

    class Meta:
        identifier = "template"



class Pod(ArnheimModel):
    template: Optional[Template]
    id: int
    status: Optional[str]

    class Meta:
        identifier = "pod"


class Provision(ArnheimModel):
    pod: Optional[Pod]
    node: Optional[Node]
    status: Optional[str]
    statusmessage: Optional[str]
    reference: Optional[str]

    class Meta:
        identifier = "provision"


class AssignationStatus(str, Enum):
    ERROR = "ERROR"
    PROGRESS = "PROGRESS"
    DEBUG = "DEBUG"
    DONE = "DONE"
    CRITICAL ="CRITICAL"
    PENDING = "PENDING"


class ProvisionStatus(str, Enum):
    ERROR = "ERROR"
    PROGRESS = "PROGRESS"
    DEBUG = "DEBUG"
    ASSIGNED = "ASSIGNED"
    CRITICAL ="CRITICAL"
    PENDING = "PENDING"

class PodStatus(str, Enum):
    PENDING = "PENDING"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"

class Assignation(ArnheimModel):
    pod: Optional[Pod]
    id: Optional[int]
    inputs: Optional[dict]
    outputs: Optional[dict]
    message: Optional[str]
    status: Optional[str]
    statusmessage: Optional[str]

    class Meta:
        identifier = "assignation"


class VartPod(Pod):
    volunteer:  Optional[Volunteer] 

    class Meta:
        identifier = "vartpod"




Avatar.update_forward_refs()
DataPoint.update_forward_refs()
Node.update_forward_refs()



from bergen.types.node.widgets.base import BaseWidget


class BasePort:

  def __init__(self, type, key=None, label=None, widget=None, description=None, required=True, primary=False) -> None:
      self.type = type
      self.key = key
      self.label = label
      self.widget = widget
      assert self.widget is None or isinstance(self.widget, BaseWidget), "Widget must be either none or instance of BaseWidget"
      self.description = description
      self.required = required
      self.primary = primary
      super().__init__()


  def serialize(self):
      assert self.key is not None, "Please provide at least a key to your Port"
      widgetFragment = {"widget": self.widget.serialize()} if self.widget is not None else {}
      return {
        "type": self.type,
        "key": self.key,
        "label" : self.label or self.key.capitalize(),
        "description": self.description,
        "required": self.required,
        "primary": self.primary,
        **widgetFragment
      }


  def __call__(self, key):
      self.key = key
      return self


from bergen.query import DelayedGQL


NODE_QUERY = DelayedGQL("""
query Node($id: ID!){
  node(id: $id){
    id
    name
    image
    inputs {
      __typename
      key
      required
      ... on ModelPortType {
        identifier
      }
    }
    outputs {
      __typename
      key
      required
      ... on ModelPortType {
        identifier
      }
    }
  }
}
""")


NODE_FILTER_QUERY = DelayedGQL("""
query NodeFilter($name: String){
  nodes(name: $name){
    id
    name
    repository {
      name
    }
    description
  }
}
""")


CREATE_NODE_MUTATION = DelayedGQL("""
  mutation CreateNodeMutation($description: String!, $inputs: [PortInputType]!, $outputs: [PortInputType]!, $package: String!, $interface: String!, $name: String!){
  createNode(description: $description, inputs: $inputs, outputs: $outputs, package:$package, interface: $interface, name: $name){
    id
    name
    image
    inputs {
      __typename
      key
      required
      ... on ModelPortType {
        identifier
      }
    }
    outputs {
      __typename
      key
      required
      ... on ModelPortType {
        identifier
      }
    }
  }
} 
""")


from bergen.wards.graphql.default import GraphQlWard
from bergen.auths.base import BaseAuthBackend
from typing import Callable
from bergen.enums import DataPointType
from bergen.clients.base import BaseWard
import logging

logger = logging.getLogger(__name__)

datapointregistry = None

class DataPointRegistry(object):


    def __init__(self) -> None:
        self.pointNameWardMap: dict[str, BaseWard] = {}
        self.builders =  {
                # Default Builders for standard
                DataPointType.GRAPHQL: lambda datapoint, auth: GraphQlWard(host=datapoint.host, port=datapoint.port, token=auth.getToken(), protocol=auth.getProtocol())
        }

    def registerClientBuilder(self, type:str , builder: Callable):
        self.builders[type] = builder

    def getClientForData(self,point, auth: BaseAuthBackend) -> BaseWard:
        if point.name in self.pointNameWardMap:
            return self.pointNameWardMap[point.name]

        logger.info("Creating new DataPoint parser")

        if point.type in self.builders:
            builder = self.builders[point.type]
            self.pointNameWardMap[point.name] = builder(point, auth)
            return self.pointNameWardMap[point.name]
        else:
            raise NotImplementedError("We have no idea how to build that datatype")






def get_datapoint_registry() -> DataPointRegistry:
    global datapointregistry
    if datapointregistry is None:
        datapointregistry = DataPointRegistry()
    return datapointregistry


from abc import ABC
from bergen.constants import END_GQL, MARK_GQL, YIELD_GQL
from bergen.schema import Volunteer, Assignation, AssignationStatus
import logging


logger = logging.getLogger(__name__)

class PodHelper(ABC):
    ''' This helps your pod do access the arnheim instance '''


    def __init__(self, volunteer: Volunteer, assignation: Assignation, arnheim):
        self.volunteer = volunteer
        self.assignation = assignation
        self.arnheim = arnheim


    def update(self, message, level=logging.DEBUG):
        if level == logging.DEBUG:
            MARK_GQL.run(ward=self.arnheim.main_ward, variables={"message": message, "assignation": self.assignation.id, "level": AssignationStatus.PROGRESS})
        else:
            return

    def end(self, outputs: dict):
        return END_GQL.run(ward=self.arnheim.main_ward, variables={"outputs": outputs, "assignation": self.assignation.id})

    def nudge(self, outputs: dict):
        return YIELD_GQL.run(ward=self.arnheim.main_ward, variables={"outputs": outputs, "assignation": self.assignation.id})

    def critical(self, message):
        return MARK_GQL.run(ward=self.arnheim.main_ward, variables={"message": message, "assignation": self.assignation.id, "level": AssignationStatus.CRITICAL})





from abc import ABC
import re
from bergen.wards.base import BaseWard
from bergen.query import GQL, TypedGQL
import requests


class GraphQlException(Exception):
    pass

class ProtocolException(GraphQlException):
    pass

class SyntaxException(GraphQlException):
    pass

class QueryException(GraphQlException):
    pass


class BaseGraphQLWard(BaseWard, ABC):
    can_subscribe = False
    

    def __init__(self, port=8000, host="localhost", protocol="http", token=None) -> None:
        self._graphql_endpoint = f"{protocol}://{host}:{port}/graphql"
        self._token = token
        self._headers = {"Authorization": f"Bearer: {self._token}"}
        super().__init__(port=port, host=host, protocol=protocol, token=token)


    def post(self, json, url = None, headers = None):
        request = requests.post(url or self._graphql_endpoint, json=json, headers= headers or self._headers)
        if request.status_code == 200:
            result = request.json()
            if "errors" in result:
                raise QueryException(",".join([error["message"] for error in result["errors"]]))
            return result

        if request.status_code == 400:
            result = request.json()
            if "errors" in result:
                raise QueryException(",".join([error["message"] for error in result["errors"]]))
                
        else:
            raise ProtocolException(f"Post failed to run by returning code of {request.status_code}. {request.content}")

    def run(self, gql: TypedGQL, variables: dict):
        result = self.post(gql.combine(variables))
        return gql.extract(result)


from bergen.logging import setLogging
from bergen.auths.base import BaseAuthBackend
from bergen.wards.base import BaseWard

import logging


logger = logging.getLogger(__name__)
import os


class BaseBergen:


    def __init__(self, auth: BaseAuthBackend, main_ward: BaseWard, auto_negotiate=True, bind=True, log=logging.INFO, local=None, **kwargs) -> None:
        
        if bind: 
            # We only import this here for typehints
            from bergen.registries.arnheim import set_current_arnheim
            set_current_arnheim(self)


        if log is not False:
            print(r"     _               _          _              ____ _ _            _    ")   
            print(r"    / \   _ __ _ __ | |__   ___(_)_ __ ___    / ___| (_) ___ _ __ | |_  ")
            print(r"   / _ \ | '__| '_ \| '_ \ / _ \ | '_ ` _ \  | |   | | |/ _ \ '_ \| __| ")
            print(r"  / ___ \| |  | | | | | | |  __/ | | | | | | | |___| | |  __/ | | | |_  ")
            print(r" /_/   \_\_|  |_| |_|_| |_|\___|_|_| |_| |_|  \____|_|_|\___|_| |_|\__| ")
            print(r"")
            setLogging(True, log)


        self.local = local if local is not None else os.getenv("ARNHEIM_LOCAL") == "1" 
        if self.local:
            logger.info("Running in Local Mode")
        self.auth = auth
        self.main_ward = main_ward
        self._transcript = None
        self.identifierDataPointMap = {}
        self.identifierWardMap: dict[str, BaseWard] = {}
        super().__init__()

        if auto_negotiate == True:
            self.negotiate()         

    @property
    def transcript(self):
        assert self._transcript is not None, "We have to negotiate first with our"
        return self._transcript


    def getExtensionSettings(self, extension):
        assert extension in self.transcript.extensions, f"Arnheim seems to have no idea about this Extension {extension}"
        return self.transcript.extensions[extension]


    def getWardForIdentifier(self, identifier):
        assert self._transcript is not None, "We cannot get query Identifiers on Datapoint before having negotiated them"
        if identifier in self.identifierWardMap:
            return self.identifierWardMap[identifier]
        else:
            return self.main_ward
            



    def negotiate(self, clientType = None):
        
        from bergen.constants import NEGOTIATION_GQL

        self._transcript = NEGOTIATION_GQL.run(ward=self.main_ward, variables={"clientType": clientType or self.auth.getClientType(), "local": self.local})
        logger.info(f"Successfully Got Protocols")
        
        assert self._transcript.models is not None, "We apparently didnt't get any points"

        from bergen.registries.datapoint import get_datapoint_registry
        datapoint_registry = get_datapoint_registry()

        self.identifierDataPointMap = {model.identifier.lower(): model.point for model in self._transcript.models}
        self.identifierWardMap = {model.identifier.lower(): datapoint_registry.getClientForData(model.point, self.auth) for model in self._transcript.models}

        logger.info("We established all Proper Clients for the interaction!")
        logger.info("Arnheim Ready!!!!!!!!")

    def getWard(self):
        return self.main_ward

    def _repr_html_(self):
        return f"""
            <p> Arnheim Client <p>
            <table>
                <tr>
                    <td> Connected to </td> <td> {self.main_ward.host} </td>
                </tr>
            </table>



        """



from abc import ABC, abstractmethod
from bergen.query import TypedGQL
from typing import Callable, Dict, Generic, List, TypeVar, Any



ModelType = TypeVar("ModelType")

class ModelConfigurationError(Exception):
    pass

class ModelManager(Generic[ModelType]):

    def __call__(self, model: ModelType, meta) -> Any:
        self.model = model
        self.meta = meta
        return self

    def get_ward(self):
        try:
            identifier = self.model.Meta.identifier
        except Exception as e:
            raise ModelConfigurationError(f"Make soure your Model {self.model.__name__}overwrites Meta identifier: {e}")
        from bergen.registries.arnheim import get_current_arnheim


        return get_current_arnheim().getWardForIdentifier(identifier=identifier)


    def _call_meta(self, attribute, ward=None, **kwargs):
        method =  getattr(self.meta, attribute, None)
        assert method is not None, f"Please provide the {attribute} parameter in your ArnheimModel meta class for SchemaClass {self.model.__name__} "
        typed_gql: TypedGQL = method(self.model)    
        return typed_gql.run(ward=ward, variables=kwargs)


    def get(self, ward=None, **kwargs) -> ModelType:
        return self._call_meta("get", ward=ward, **kwargs)
        
    def create(self, ward=None, **kwargs) -> ModelType:
        return self._call_meta("create", ward=ward, **kwargs)

    def filter(self, ward=None, **kwargs) -> List[ModelType]:
        return self._call_meta("filter", ward=ward, **kwargs)

    def all(self, ward=None) -> List[ModelType]:
        return self._call_meta("filter", ward=ward)




from typing import TypeVar
from bergen.wards.base import BaseWard
from bergen.query import TypedGQL

T = TypeVar("T")

class SubscribeMixIn:


    def subscribe(self, gql: TypedGQL[T], variables: dict= {}, force_endpoint=None) -> T:
        assert gql.type == "subscription", "This is not a mutation!"
        identifier = force_endpoint or gql.cls.Meta.identifier
        ward: BaseWard = self.getClientForIdentifier(identifier)
        return ward.subscribe(gql, variables=variables)




from bergen.types.node.ports.base import BasePort
from bergen.types.model import ArnheimModel

class ModelPort(BasePort):

  def __init__(self, modelClass: ArnheimModel, **kwargs) -> None:
      self.modelClass = modelClass
      super().__init__("model",**kwargs)

  def serialize(self):
      return {**super().serialize(),"identifier" : self.modelClass.getMeta().identifier}


from bergen.wards.graphql.base import QueryException
from bergen.query import TypedGQL
import json
from bergen.wards.graphql.default import GraphQlWard
import logging
import websockets
import uuid

logger = logging.getLogger(__name__)

class SubscriptionGraphQLWard(GraphQlWard):
    subprotocols = ["graphql-ws"]
    extra_headers = {}
    can_subscribe = True

    def __init__(self, port=8000, host="localhost", protocol="http", token=None) -> None:
        self._websocket_endpoint = f"ws://{host}:{port}/graphql?token={token}" 
        super().__init__(port=port, host=host, protocol=protocol, token=token)


    def build_ws_message(self, type, payload = None, id = None):
        message = {"type": type, "payload": payload or {}}
        if id: message.update({"id":id})
        return json.dumps(message)


    async def subscribe(self,gql: TypedGQL, variables: dict, extract=True):
        logger.info(f"Starting Subscription for: {gql.operation_name}")
        async with websockets.connect(self._websocket_endpoint, subprotocols=self.subprotocols, extra_headers=self.extra_headers) as websocket:
            await websocket.send(self.build_ws_message("connection_init"))

            id = str(uuid.uuid4())
            payload = gql.combine(variables=variables)
            await websocket.send(self.build_ws_message("start", payload=payload, id=id))

            while True:
                signal = await websocket.recv()
                answer = json.loads(signal)
                type = answer["type"]
                if type == "ka": pass # We received a keep alive signal
                if type == "complete": break
                if type == "data":
                    payload = answer["payload"]
                    if "errors" in payload: raise QueryException(f'Error in Payload: {payload["errors"]}')
                    test = gql.extract(payload)
                    yield gql.cls(**test)


from bergen.registries.matcher import get_current_matcher
from bergen.enums import TYPENAMES
from bergen.schema import Node


class ExpansionError(Exception):
    pass



def expandInputs(node: Node, inputs: dict) -> dict:

    #assert node.inputs is not None, "Your Query for Nodes seems to not provide any field for inputs, please use that in your get statement"
    #assert len(node.inputs) > 0  is not None, "Your Node seems to not provide any inputs, calling is redundant"

    kwargs = {}
    for port in node.inputs:
        if port.key not in inputs:
            if port.required:
                raise ExpansionError(f"We couldn't expand {port.key} because it wasn't provided by our Inputs, wrong assignation!!!")
            else:
                break

        if port.TYPENAME == TYPENAMES.MODELPORTTYPE:
            modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
            instance =  modelClass.objects.get(id=inputs[port.key])
            kwargs[port.key] = instance
        else:
            kwargs[port.key] = inputs[port.key]

    return kwargs


def expandOutputs(node: Node, inputs: dict) -> dict:

    kwargs = {}
    for port in node.outputs:
        if port.TYPENAME == TYPENAMES.MODELPORTTYPE:
            modelClass = get_current_matcher().getModelForIdentifier(identifier=port.identifier)
            instance =  modelClass.objects.get(id=inputs[port.key])
            kwargs[port.key] = instance
        else:
            kwargs[port.key] = inputs[port.key]

    return kwargs







import inspect
from bergen.utils import ExpansionError, expandInputs
from aiostream.aiter_utils import async_
from bergen.constants import HOST_GQL, QUEUE_GQL
from bergen.helpers.pod import PodHelper
from bergen.helpers.host import HostHelper
from bergen.schema import Assignation, Node
import logging
import asyncio
from aiostream import stream, pipe
logger = logging.getLogger(__name__)

class HostMixIn:


    def __init__(self, name="test") -> None:
        self.name = name
        self.hostHelper = HostHelper(self) #

        super().__init__()


    def template(self, node: Node, version="standard", threaded=False):
        """Template is a decorator for marking a function as a Template

        Args:
            node (Node): The Node that you want to template for
            version (str, optional): Do you want to specify the version. Defaults to "standard".
            threaded (bool, optional): Should we run a pod of this template in a seperate thread?. Defaults to False.

        Returns:
            void
        """

        logger.info(f"Registering Node for {node.name}")
        volunteer = self.hostHelper.volunteer(node, version=version)

        def real_decorator(function):

            async def wrapper(assignation: Assignation):
                helper = PodHelper(volunteer, assignation, self)
                logger.info(f"Called with inputs {assignation.inputs}")

                try:
                    kwargs = expandInputs(node, assignation.inputs)
                except ExpansionError as e:
                    logger.error(e)
                    helper.critical(str(e))
                    return
                
                try:


                    if inspect.isasyncgenfunction(function):
                        # Exicting this is a generator who will yield results
                        yieldedresult = None
                        async for result in function(helper, **kwargs):
                            logger.info(f"Yielding {result}")
                            helper.nudge(result)
                            yieldedresult = result
                        
                        helper.end(yieldedresult)

                        return

                    if inspect.iscoroutinefunction(function):
                        result = await function(helper, **kwargs)
                        helper.end(result)

                        return


                    elif threaded:
                        # Threaded functions
                        threadhelper = None 
                        try: 
                            threadhelper = asyncio.to_thread
                        except:
                            logger.warn("Threading does not work below Python 3.9")
                        
                        if threadhelper:
                            result = await threadhelper(function, helper, **kwargs)
                        else:
                            result = function(helper, **kwargs)
                            helper.end(result)

                        return
                    else:
                        # We are dealing with a normal function. boring
                        result = function(helper, **assignation.inputs)
                        helper.end(result)

                        return
                    
                except Exception as e:
                    logger.error(e)
                    helper.critical(str(e))

                
                return 


            wrapper.__name__ = function.__name__
            self.hostHelper.addVolunteerFunction(volunteer, wrapper)

            return function


        return real_decorator


    async def host(self):

        volunteers = self.hostHelper.getVolunteers()

        queued = stream.merge(*[QUEUE_GQL.subscribe(ward=self.main_ward, variables= {"id": volunteer.id}) for volunteer in volunteers])
    
        assigned_pods = (queued 
            | pipe.action(lambda pod: self.hostHelper.addPod(pod)) # We are adding the Pod to the procedure
            | pipe.action(lambda pod: logger.warn(f"Got promoted to Host {pod}"))
        )
        
        assigned_assignations = (assigned_pods
             | pipe.flatmap(lambda pod: HOST_GQL.subscribe(ward=self.main_ward, variables={"pod": pod.id}))
             | pipe.action(lambda assignation: logger.warn(f"We are assigning this to our {assignation.inputs}"))
             | pipe.map(async_(lambda assignation: self.hostHelper.getFunctionForPod(assignation.pod)(assignation)))
             | pipe.map(lambda result: logger.info(f"Assignation finished with {result}"))
        )

        async with assigned_assignations.stream() as streamer:
            async for item in streamer:
                print(item)

        return True


    

    def run(self):
        logger.warn("Waiting for incoming Messages")

        loop = asyncio.get_event_loop()
        asyncio.run(self.host())







from bergen.auths.base import AuthError, BaseAuthBackend
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session 
from bergen.enums import ClientType
import logging
import time

logger = logging.getLogger(__name__)


class ArnheimBackendOauth(BaseAuthBackend):
    failedTries = 5
    auto_retry = True
    tokenurl_appendix = "o/token/"


    def __init__(self, host: str, port: int, client_id: str, client_secret: str, protocol="http", verify=True) -> None:
        self.host = host
        self.port = port
        self.client_id = client_id
        self.client_secret = client_secret
        self.protocol = protocol
        self.verify = verify

        self.base_url = f"{protocol}://{host}:{port}"       
        self.token_url = self.base_url + "/" + self.tokenurl_appendix
        
        super().__init__()

    def getToken(self) -> str:
        logger.info(f"Connecting to Arnheim at {self.token_url}")
        
        auth_client = BackendApplicationClient(client_id=self.client_id)
        oauth_session = OAuth2Session(client=auth_client)


        def fetch_token(thetry=0):
            try:
                return oauth_session.fetch_token(token_url=self.token_url, client_id=self.client_id,
                client_secret=self.client_secret, verify=self.verify)["access_token"]
            except Exception as e:
                if thetry == self.failedTries or not self.auto_retry: raise AuthError(f"Cannot connect to Arnheim instance on {self.token_url}: {e}")
                logger.error(f"Couldn't connect to the Arnheim Instance at {self.token_url}. Retrying in 2 Seconds")
                time.sleep(2)
                fetch_token(thetry=thetry + 1)

        return fetch_token()

    def getClientType(self) -> ClientType:
        return ClientType.EXTERNAL.value


    def getProtocol(self):
        return self.protocol


import logging

def setLogging(param: bool, log_level= logging.INFO):
    if param:
        LOGFORMAT = "  s%(levelname)-8s%(reset)s | %(log_color)s%(message)s%(reset)s"
        stream = logging.StreamHandler()
        logging.root.setLevel(log_level)
        stream.setLevel(log_level)

        try:
            from colorlog import ColoredFormatter
            LOGFORMAT = "  %(log_color)s%(levelname)-8s%(reset)s | %(log_color)s%(message)s%(reset)s"
            formatter = ColoredFormatter(LOGFORMAT)
            stream.setFormatter(formatter)
        except:
            pass

        logger = logging.getLogger()
        logger.setLevel(log_level)
        logger.addHandler(stream)



from bergen.wards.graphql.subscription import SubscriptionGraphQLWard
from bergen.auths.backend_application import ArnheimBackendOauth
from bergen.clients.base import BaseBergen
from bergen.clients.mixins.querymixin import QueryMixIn
from bergen.clients.mixins.subscribemixin import SubscribeMixIn
import os

class Bergen(BaseBergen, QueryMixIn, SubscribeMixIn):

    def __init__(self, host: str = "localhost", port: int = 8000, client_id: str = None, client_secret: str = None, protocol="http", bind=True, allow_insecure=None, is_local=None, **kwargs) -> None:

        
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1" if allow_insecure is not None else os.getenv("OAUTHLIB_INSECURE_TRANSPORT", "0")
        os.environ["ARNHEIM_LOCAL"] = "0" if is_local is not None else os.getenv("ARNHEIM_LOCAL", "0")


        auth = ArnheimBackendOauth(host=host, port=port, client_id=client_id, client_secret=client_secret, protocol="http", verify=True)
        main_ward = SubscriptionGraphQLWard(host=host, port=port, protocol=protocol, token=auth.getToken())

        super().__init__(auth, main_ward, auto_negotiate=True, bind=bind, **kwargs)


from abc import ABC, abstractmethod
from bergen.types.manager import ModelManager
from typing import Callable, Dict, Generic, List, TypeVar, Type
from pydantic.fields import Field
from pydantic.main import BaseModel, ModelMetaclass
from bergen.query import DelayedGQL, GQL, TypedGQL

class ArnheimModelException(Exception):
    pass

class ModelDoesNotExistError(ArnheimModelException):
    pass

class ArnheimModelConfigurationError(ArnheimModelException):
    pass


ModelType = TypeVar("ModelType", bound="ArnheimModel", covariant=True)

class ArnheimModelManager(ABC, Generic[ModelType]):

    def __init__(self, model: ModelType, meta) -> None:
        self.model = model
        self.meta = meta
        super().__init__()

    def get_ward(self):
        try:
            identifier = self.model.Meta.identifier
        except Exception as e:
            raise ArnheimModelConfigurationError(f"Make soure your Model {self.model.__name__}overwrites Meta identifier: {e}")
        from bergen.registries.arnheim import get_current_arnheim


        return get_current_arnheim().getWardForIdentifier(identifier=identifier)


    def _call_meta(self, attribute, ward=None, **kwargs):
        method =  getattr(self.meta, attribute, None)
        assert method is not None, "Please provide the {attribute} parameter in your ArnheimModel meta class "
        typed_gql: TypedGQL = method(self.model)    
        return typed_gql.run(ward=ward, variables=kwargs)


    def get(self, ward=None, **kwargs) -> ModelType:
        return self._call_meta("get", ward=ward, **kwargs)
        
    def create(self, ward=None, **kwargs) -> ModelType:
        return self._call_meta("create", ward=ward, **kwargs)

    def filter(self, ward=None, **kwargs) -> List[ModelType]:
        return self._call_meta("filter", ward=ward, **kwargs)

    def all(self, ward=None):
        return self._call_meta("filter", ward=ward)



class ArnheimModelMeta(ModelMetaclass):

    def __new__(mcls, name, bases, attrs):
        
        slots = set(attrs.pop('__slots__', tuple())) # The slots from: https://github.com/samuelcolvin/pydantic/issues/655#issuecomment-610900376
        for base in bases:
            if hasattr(base, '__slots__'):
                slots.update(base.__slots__)

        if '__dict__' in slots:
            slots.remove('__dict__')
        attrs['__slots__'] = tuple(slots)

        mcls.overriden_manager = attrs.pop("objects") if "objects" in attrs else None
        return super(ArnheimModelMeta, mcls).__new__(mcls, name, bases, attrs)


    @property
    def objects(cls: Type[ModelType]) -> ArnheimModelManager[ModelType]:
        return cls.__objects


    def __init__(self, name, bases, attrs):
        super(ArnheimModelMeta, self).__init__(name, bases, attrs)
        if attrs["__qualname__"] != "ArnheimModel":
            # This gets allso called for our Baseclass which is abstract
            meta = attrs["Meta"] if "Meta" in attrs else None
            assert meta is not None, "Please provide a Meta class in your Arnheim Model"
            
            managerClass = self.overriden_manager
            if managerClass:
                self.__objects = managerClass(self, meta)
            else:
                self.__objects = ArnheimModelManager[ModelType](self, meta)


            
            overwriteType = getattr(meta, "overwrite_default", False)
            register = getattr(meta, "register", True)
            if register:
                from bergen.registries.matcher import get_current_matcher
                identifier = getattr(meta, "identifier", None)
                assert identifier is not None, "Please provide identifier in your Meta class to register the Model, or specifiy register==False"
                get_current_matcher().registerModelForIdentifier(identifier, self, overwrite=overwriteType)






class ArnheimModel(BaseModel, metaclass=ArnheimModelMeta):
    TYPENAME: str = Field(None, alias='__typename')


    @classmethod
    def getMeta(cls):
        return cls.Meta

    def __repr__(self) -> str:
        from pprint import pformat
        return pformat(self.__dict__, indent=4, width=1)

    def _repr_html_(self):
        
        def buildTable(attributes: Dict):
            tablestring = "<table>"
            for key, value in attributes.items():
                tablestring = tablestring + (f"""
                    <tr>
                        <td>{key.capitalize()}</td>
                        <td>{value}</td>
                    </tr>
                """)
            return tablestring + "</table>"



        return f"""
            <p> Instance of {self.__class__.__name__} <p>
            {buildTable(self.__dict__)}
        """

    def _repr_list_(self):
        string = f"{self.__class__.__name__}("
        if self.id is not None: string += str(self.id) + ", "
        if self.name is not None: string += str(self.name) +", "
        string = string[:-2] + ")"
        return string


    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            super().__setattr__(attr, value)



from typing import Generic
from bergen.query import  TypedGQL
from typing import TypeVar


T = TypeVar("T")

class BaseWard(Generic[T]):

    def __init__(self, port=8000, host="localhost", protocol="http", token=None) -> None:
        self.port = port
        self.host = host
        self.protocol = protocol
        self.token = token
        super().__init__()


    def run(self, gql: TypedGQL, variables: dict):
        return gql.cls(**{})


from bergen.wards.graphql.subscription import SubscriptionGraphQLWard
from bergen.registries.arnheim import get_current_arnheim
import asyncio
from bergen.types.model import ArnheimModel
from bergen.extenders.base import BaseExtender
import logging
from aiostream import stream
import uuid



logger = logging.getLogger(__name__)



def serialize(inputs: dict):
    serialized_inputs = {}
    
    for key, value in inputs.items():
        if isinstance(value, ArnheimModel):
            serialized_inputs[key] = value.id
        else:
            serialized_inputs[key] = value

    return serialized_inputs


class AsyncPodHandler:

    def __init__(self, provisionHandler, ward, node, pod) -> None:
        self._pod = pod
        self._node = node
        self._provisionHandler = provisionHandler
        self._ward = ward
        super().__init__()


    async def stream(self, inputs: dict):
        from arnheim.constants import ASSIGN_GQL
        logger.info(f"Trying to assign to {self._pod}")
        serialized_inputs = {}

        serialized_inputs = serialize(inputs)

        expandedoutputs = None
        async for item in ASSIGN_GQL.subscribe(self._ward, variables={"pod": self._pod.id, "inputs": serialized_inputs, "reference": str(uuid.uuid4())}):
            if item.status == "DONE":
                from arnheim.utils import expandOutputs
                expandedoutputs = expandOutputs(self._node, item.outputs)
                yield expandedoutputs
                break
            elif item.status == "YIELD":
                from arnheim.utils import expandOutputs
                expandedoutputs = expandOutputs(self._node, item.outputs)
                yield expandedoutputs
            elif item.status == "CRITICAL":
                raise Exception(item.statusmessage)
            else:
                pass


    async def assign(self, inputs: dict):
        from arnheim.constants import ASSIGN_GQL
        logger.debug(f"Trying to assign to {self._pod}")
        serialized_inputs = {}

        serialized_inputs = serialize(inputs)

        async for item in ASSIGN_GQL.subscribe(self._ward, variables={"pod": self._pod.id, "inputs": serialized_inputs, "reference": str(uuid.uuid4())}):
            logger.info(f"Received {item}")
            if item.status == "DONE":
                from arnheim.utils import expandOutputs
                expandedoutputs = expandOutputs(self._node, item.outputs)
                return expandedoutputs
            elif item.status == "CRITICAL":
                raise Exception(item.statusmessage)
            else:
                pass




        

    async def kill(self):
        await asyncio.sleep(0.01)
        logger.info("Closed Connection")


class ProvisionHandler:

    def __init__(self, nodeModel, ward: SubscriptionGraphQLWard, provider="vart") -> None:
        super().__init__()
        self._pod = None
        self._ward = ward
        self._provider = provider
        self._node = nodeModel

    async def __aenter__(self):
        from arnheim.constants import PROVIDE_GQL

        async for item in PROVIDE_GQL.subscribe(self._ward, variables={"node": self._node.id, "selector": {"provider": self._provider, "kwargs": {"volunteer": 2}}, "reference": str(uuid.uuid4())}):
            if item.pod.status == "ACTIVE":
                self._provisionreference = item.reference 
                self._pod = AsyncPodHandler(self, self._ward, self._node, item.pod)
                return self._pod
            else:
                pass

    


    async def assign(self, inputs):
        async with self as handler:
            return await handler.assign(inputs)

        
    async def __aexit__(self, exc_type, exc, tb):
        await self._pod.kill()
        print(exc_type, exc, tb)
        self._pod = None


class NodeExtender(BaseExtender):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        self._pod = "Hallo"

    def __call__as_stream(self, inputs, provider):
        logger.info(f"Called as Stream")
        return stream.count(interval=0.2)

    async def __call_async(self, inputs, provider):
        
        async with self.provide(provider=provider) as handler:
            return await handler.assign(inputs=inputs)

    async def __await_call_async(self, inputs,provider):
        return await self.__call_async(inputs, provider)

    def __call_sync(self, inputs, provider):
        logger.info(f"Called in Sync")
        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(self.__await_call_async(inputs, provider))
        loop.close()
        return result

    def __call__(self, inputs: dict, provider="vart", stream=False):
        """Call this method to notify all subscriptions in the group.
        This method can be called from both synchronous and asynchronous
        contexts. If you call it from the asynchronous context then you
        have to `await`.
        Args:
            group: Name of the subscription group which members must be
                notified. `None` means that all the subscriptions of
                type will be triggered.
            payload: The payload delivered to the `publish` handler.
                NOTE: The `payload` is serialized before sending
                to the subscription group.
        """
        try:
            event_loop = asyncio.get_event_loop()
        except RuntimeError:
            pass
        else:
            if event_loop.is_running():
                if stream: return self.__call__as_stream(inputs, provider)
                return self.__call_async(inputs, provider)
        

            
        return self.__call_sync(inputs, provider)

    def provide(self, provider = "vart"):
        ward = get_current_arnheim().getWard()
        self._provisionhandler = ProvisionHandler(self, ward, provider=provider)
        return self._provisionhandler


    def _repr_html_(self):
        string = f"{self.name}</br>"

        for input in self.inputs:
            string += "Inputs </br>"
            string += f"Port: {input._repr_html_()} </br>"

        for output in self.outputs:
            string += "Outputs </br>"
            string += f"Port: {output._repr_html_()} </br>"


        return string


from bergen.clients.mixins.hostmixin import HostMixIn
from bergen.wards.graphql.subscription import SubscriptionGraphQLWard
from bergen.auths.backend_application import ArnheimBackendOauth
from bergen.clients.base import BaseArnheim
from bergen.clients.mixins.querymixin import QueryMixIn
from bergen.clients.mixins.subscribemixin import SubscribeMixIn


class ArnheimHost(BaseArnheim, QueryMixIn, SubscribeMixIn,HostMixIn ):

    def __init__(self, host: str = "localhost", port: int = 8000, client_id: str = None, client_secret: str = None, protocol="http", bind=True, log=True, name="") -> None:

        auth = ArnheimBackendOauth(host=host, port=port, client_id=client_id, client_secret=client_secret, protocol="http", verify=True)

        main_ward = SubscriptionGraphQLWard(host=host, port=port, protocol=protocol, token=auth.getToken())

        super().__init__(auth, main_ward, auto_negotiate=True, bind=bind, log=log)


from bergen.clients.base import BaseBergen

CURRENT_ARNHEIM = None

def get_current_arnheim() -> BaseBergen:
    global CURRENT_ARNHEIM
    if CURRENT_ARNHEIM is None:
        raise Exception("No Client was initialized before")
    else:
        return CURRENT_ARNHEIM


def set_current_arnheim(arnheim):
    global CURRENT_ARNHEIM
    CURRENT_ARNHEIM = arnheim



from typing import TypeVar
from bergen.query import TypedGQL
from bergen.wards.base import BaseWard


T = TypeVar("T")



class QueryMixIn:


    def query(self, gql: TypedGQL[T], variables: dict= {}, force_endpoint=None) -> T:
        assert gql.type == "query", "This is not a query!"
        identifier = force_endpoint or gql.cls.Meta.identifier
        ward: BaseWard = self.getClientForIdentifier(identifier)
        return gql.run(gql, variables=variables, ward=ward)


    def mutate(self, gql: TypedGQL[T], variables: dict= {}, force_endpoint=None) -> T:
        assert gql.type == "mutation", "This is not a mutation!"
        identifier = force_endpoint or gql.cls.Meta.identifier  
        ward: BaseWard = self.getClientForIdentifier(identifier)
        return gql.run(gql, variables=variables, ward=ward)
